from __future__ import annotations
import logging
from typing import Any, List, Optional, Sequence, cast
from starlette.applications import Starlette
from starlette.middleware.sessions import SessionMiddleware
from starlette.routing import Mount, Route
from starlette.datastructures import URLPath
from starlette.types import Receive, Scope, Send
from starlette.templating import Jinja2Templates
from z8ter import get_templates
from z8ter.route_builders import (
    build_routes_from_pages,
    build_routes_from_apis,
    build_file_routes,
)
logger = logging.getLogger("z8ter")


class Z8ter:
    def __init__(
        self,
        *,
        debug: Optional[bool] = None,
        mode: Optional[str] = None,
        routes: Optional[Sequence[Route | Mount]] = None,
        sessions: bool = False,
        session_secret: Optional[str] = None,
    ) -> None:
        self._extra_routes: List[Route | Mount] = list(routes or [])
        self.mode: str = (mode or "prod").lower()
        self.debug: bool = False
        if debug is None:
            self.debug = bool(self.mode == "dev")
        else:
            self.debug = bool(debug)
        self.app: Starlette = Starlette(debug=self.debug,
                                        routes=self._assemble_routes())

        if sessions:
            if session_secret:
                secret: str = session_secret
            else:
                raise ValueError(
                    "Z8ter: session_secret is required when sessions=True."
                )
            self.app.add_middleware(SessionMiddleware, secret_key=secret)

        def _url_for(name: str, filename: Optional[str] = None,
                     **params: Any) -> str:
            if filename is not None:
                params["path"] = filename
            path: URLPath = self.app.url_path_for(name, **params)
            return str(path)

        templates: Jinja2Templates = get_templates()
        templates.env.globals["url_for"] = _url_for

        if self.debug:
            logger.warning("🧪 Z8ter running in DEBUG mode")

    async def __call__(
            self, scope: Scope, receive: Receive, send: Send
            ) -> None:
        await self.app(scope, receive, send)

    def _assemble_routes(self) -> List[Route | Mount]:
        routes: List[Route | Mount] = []
        routes += self._extra_routes
        routes.append(build_file_routes())
        routes += build_routes_from_pages()
        routes += build_routes_from_apis()
        return routes

    def _ensure_services_registry(self) -> None:
        state: Any = self.app.state
        if not hasattr(state, "services"):
            state.services = {}

    def add_service(self, obj: object, *, replace: bool = False) -> str:
        """
        Registers a process-wide service under app.state.
        Access via:
        request.app.state.<name> or
        request.app.state.services[name]
        """
        self._ensure_services_registry()
        key: str = (obj.__class__.__name__).rstrip("_").lower()

        state: Any = self.app.state
        services: dict[str, object] = cast(dict[str, object], state.services)

        if key in services and not replace:
            raise ValueError(f"Service '{key}' already exists." +
                             "Use replace=True to overwrite.")

        services[key] = obj
        setattr(state, key, obj)
        return key



import logging
import asyncio


class IgnoreCancelledFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        if record.exc_info:
            return not isinstance(record.exc_info[1], asyncio.CancelledError)
        return "CancelledError" not in record.getMessage()


def uvicorn_log_config(dev: bool = True) -> dict:
    return {
        "version": 1,
        "disable_existing_loggers": False,
        "filters": {
            "ignore_cancelled": {
                "()": "z8ter.logging_utils.IgnoreCancelledFilter"
            }
        },
        "formatters": {
            "plain": {"format": "%(message)s"}
        },
        "handlers": {
            "rich": {
                "class": "rich.logging.RichHandler",
                "rich_tracebacks": True,
                "markup": True,
                "show_level": True,
                "show_time": True,
                "show_path": False,
                "log_time_format": "[%X]",
                "level": "DEBUG" if dev else "INFO",
                "formatter": "plain",
                "filters": ["ignore_cancelled"],
            },
        },
        "loggers": {
            "uvicorn":        {"handlers": ["rich"], "level": "INFO",
                               "propagate": False},
            "uvicorn.error":  {"handlers": ["rich"], "level": "INFO",
                               "propagate": False},
            "uvicorn.access": {"handlers": ["rich"],
                               "level": "WARNING" if dev else "INFO",
                               "propagate": False},
            "uvicorn.lifespan": {"handlers": ["rich"],
                                 "level": "INFO", "propagate": False},
            "z8ter":         {"handlers": ["rich"],
                              "level": "DEBUG" if dev else "INFO",
                              "propagate": False},
        },
        "root": {"handlers": ["rich"], "level": "DEBUG" if dev else "INFO"},
    }



"""
z8ter.responses

Re-exports Starlette's response classes via lightweight subclasses.
Identical today, but may gain enhancements in future Z8ter releases.
"""

from __future__ import annotations

from starlette.responses import (
    Response as StarletteResponse,
    JSONResponse as StarletteJSONResponse,
    HTMLResponse as StarletteHTMLResponse,
    PlainTextResponse as StarlettePlainTextResponse,
    RedirectResponse as StarletteRedirectResponse,
    FileResponse as StarletteFileResponse,
    StreamingResponse as StarletteStreamingResponse,
)


class Response(StarletteResponse):
    """Z8ter Response (currently identical to Starlette's)."""
    pass


class JSONResponse(StarletteJSONResponse):
    """Z8ter JSONResponse (currently identical to Starlette's)."""
    pass


class HTMLResponse(StarletteHTMLResponse):
    """Z8ter HTMLResponse (currently identical to Starlette's)."""
    pass


class PlainTextResponse(StarlettePlainTextResponse):
    """Z8ter PlainTextResponse (currently identical to Starlette's)."""
    pass


class RedirectResponse(StarletteRedirectResponse):
    """Z8ter RedirectResponse (currently identical to Starlette's)."""
    pass


class FileResponse(StarletteFileResponse):
    """Z8ter FileResponse (currently identical to Starlette's)."""
    pass


class StreamingResponse(StarletteStreamingResponse):
    """Z8ter StreamingResponse (currently identical to Starlette's)."""
    pass


__all__: list[str] = [
    "Response",
    "JSONResponse",
    "HTMLResponse",
    "PlainTextResponse",
    "RedirectResponse",
    "FileResponse",
    "StreamingResponse",
]



from pathlib import Path
import argparse
import uvicorn
import socket
import logging
import z8ter
from z8ter.logging_utils import uvicorn_log_config
logger = logging.getLogger("z8ter.cli")


def set_page_content(page_name_lower, class_name) -> dict:
    content = {}
    content["template_content"] = f"""{{% extends "base.jinja" %}}
{{% block content %}}
  <h1>{class_name}</h1>
{{% endblock %}}
"""
    content["view_content"] = f"""from z8ter.page import Page
from z8ter.requests import Request
from z8ter.responses import Response


class {class_name}(Page):
    async def get(self, request: Request) -> Response:
        return self.render(request, "{page_name_lower}.jinja", {{}})
"""
    content["ts_content"] = (
        f"export default function init{class_name}(): void {{}}"
    )
    return content


def create_page(page_name: str):
    page_name_lower = page_name.lower()
    class_name = page_name.capitalize()
    template_path = z8ter.TEMPLATES_DIR / f"{page_name_lower}.jinja"
    view_path = z8ter.VIEWS_DIR / f"{page_name_lower}.py"
    ts_path = z8ter.TS_DIR / "pages" / f"{page_name_lower}.ts"
    z8ter.TEMPLATES_DIR.mkdir(parents=True, exist_ok=True)
    z8ter.VIEWS_DIR.mkdir(parents=True, exist_ok=True)
    z8ter.TS_DIR.mkdir(parents=True, exist_ok=True)
    content = set_page_content(page_name_lower, class_name)
    if not template_path.exists():
        template_path.write_text(content["template_content"], encoding="utf-8")
        print(f"Created template: {template_path}")
    else:
        print(f"Template already exists: {template_path}")
    if not view_path.exists():
        view_path.write_text(content["view_content"], encoding="utf-8")
        print(f"Created view: {view_path}")
    else:
        print(f"View already exists: {view_path}")
    if not ts_path.exists():
        ts_path.write_text(content["ts_content"], encoding="utf-8")
        print(f"Created script: {ts_path}")
    else:
        print(f"Script already exists: {ts_path}")


def create_api(api_name: str):
    api_name_lower = api_name.lower()
    class_name = api_name.capitalize()
    api_path = z8ter.API_DIR / f"{api_name_lower}.py"
    api_content = f"""from z8ter.api import API
from z8ter.requests import Request
from z8ter.responses import JSONResponse


class {class_name}(API):
    @API.endpoint("GET", "/")
    async def get_{class_name}(self, request: Request) -> JSONResponse:
        return JSONResponse({{"message": "Hello from {class_name} API!"}}, 200)
"""
    if not api_path.exists():
        api_path.write_text(api_content, encoding="utf-8")
        print(f"Created API: {api_path}")
    else:
        print(f"API already exists: {api_path}")


def new_project(project_name: str):
    print("This feature has not been implemented yet.")
    print(f"You can start your project {project_name} by cloning the repo:")
    print("https://github.com/ashesh808/Z8ter")


def run_server(
        mode: str = "prod",
        host: str = "127.0.0.1",
        port: int = 8080,
        reload: bool | None = None,
) -> None:
    def lan_ip() -> str:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            s.connect(("8.8.8.8", 80))
            return s.getsockname()[0]
        finally:
            s.close()
    isProd = mode == "prod"
    isDev = not isProd
    reload = isDev if reload is None else reload
    if mode == "WAN":
        host = "0.0.0.0"
        logger.warning(
            f"🌐 WAN: The host has been set to {host} and reload: {reload}"
        )
    elif mode == "LAN":
        host = lan_ip()
        logger.warning(
            f"🌐 LAN: The host has been set to {host} and reload: {reload}"
        )
    if mode == "prod":
        host = "0.0.0.0"
        reload = False
        logger.warning(
            f"🚀 PROD: The host has been set to {host} and reload: {reload}"
        )
    uvicorn.run(
        "main:create_app",
        factory=True,
        host=host,
        port=port,
        reload=reload,
        app_dir=str(z8ter.BASE_DIR),
        reload_dirs=[str(z8ter.BASE_DIR)],
        log_level="info",
        log_config=uvicorn_log_config(isDev),
    )


def main():
    z8ter.set_app_dir(Path.cwd())
    parser = argparse.ArgumentParser(prog="z8", description="Z8ter CLI")
    sub = parser.add_subparsers(dest="cmd", required=True)

    p_create = sub.add_parser("create_page",
                              help="Create a new page (template + view)")
    p_create.add_argument("name",
                          help="Page name (e.g., 'home' or 'app/home')")
    p_new = sub.add_parser("new",
                           help="Create a new Z8ter project")
    p_new.add_argument("project_name",
                       help="Folder name for the new project")
    p_run = sub.add_parser("run", help="Run the app (default: prod)")
    p_run.add_argument("mode", nargs="?",
                       choices=["dev", "prod", "WAN", "LAN"],
                       help="Use 'dev' for autoreload")
    args = parser.parse_args()
    if args.cmd == "create_page":
        create_page(args.name)
    elif args.cmd == "new":
        new_project(args.project_name)
    elif args.cmd == "run":
        run_server(mode=args.mode if args.mode else "prod")
    else:
        parser.print_help()



"""
z8ter.requests
Re-exports Starlette's Request class. Identical today, but may
gain enhancements in future Z8ter releases.
"""
from __future__ import annotations
from starlette.requests import Request as StarletteRequest


class Request(StarletteRequest):
    """Z8ter's Request (currently identical to Starlette's)."""
    pass


__all__: list[str] = ["Request"]



from __future__ import annotations
from pathlib import Path
import importlib
from typing import Iterable, Type
from starlette.routing import Route, Mount
from starlette.staticfiles import StaticFiles
from starlette.endpoints import HTTPEndpoint
from z8ter import STATIC_PATH
from z8ter.api import API


def _import_module_for(rel_path: Path, package_root: str) -> object:
    mod_name = str(rel_path.with_suffix("")).replace(
        "/", ".").replace("\\", ".")
    if mod_name.startswith(f"{package_root}."):
        full_name = mod_name
    else:
        full_name = f"{package_root}.{mod_name}"
    return importlib.import_module(full_name)


def _iter_page_classes(mod) -> Iterable[Type[HTTPEndpoint]]:
    from .page import Page
    for obj in vars(mod).values():
        if isinstance(obj, type) and issubclass(obj, Page) and obj is not Page:
            yield obj


def _iter_api_classes(mod) -> Iterable[Type[API]]:
    from .api import API
    for obj in vars(mod).values():
        if isinstance(obj, type) and issubclass(obj, API) and obj is not API:
            yield obj


def _url_from_file(pages_root: Path, file_path: Path) -> str:
    rel = file_path.relative_to(pages_root).with_suffix("")
    parts = rel.parts
    url = "/" + "/".join(parts)
    if url.endswith("/index"):
        url = url[:-len("/index")] or "/"
    return url or "/"


def build_routes_from_pages(pages_dir: str = "views") -> list[Route]:
    routes: list[Route] = []
    pages_root = Path(pages_dir).resolve()
    seen_paths: set[str] = set()
    for file_path in pages_root.rglob("*.py"):
        if file_path.name == "__init__.py":
            continue
        rel_to_cwd = file_path.relative_to(Path().resolve())
        mod = _import_module_for(rel_to_cwd, pages_dir)
        classes = list(_iter_page_classes(mod))
        if not classes:
            continue
        base_path = _url_from_file(pages_root, file_path)
        for cls in classes:
            path = getattr(cls, "path", None) or base_path
            if path in seen_paths and getattr(cls, "path", None) is None:
                path = f"{base_path}/{cls.__name__.lower()}"
            if path not in seen_paths:
                routes.append(Route(path, cls))
                seen_paths.add(path)
    return routes


def build_routes_from_apis(api_dir: str = "api") -> list[Mount]:
    routes: list[Mount] = []
    pages_root = Path(api_dir).resolve()
    for file_path in pages_root.rglob("*.py"):
        if file_path.name == "__init__.py":
            continue
        rel_to_cwd = file_path.relative_to(Path().resolve())
        mod = _import_module_for(rel_to_cwd, api_dir)
        classes = list(_iter_api_classes(mod))
        if not classes:
            continue
        for cls in classes:
            routes.append(cls.build_mount())
    return routes


def build_file_routes() -> Mount:
    if STATIC_PATH.exists():
        mt = Mount(
            "/static", StaticFiles(directory=str(STATIC_PATH)), name="static"
        )
    return mt
